/*

 ideservd.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ideservd.h"
#include <errno.h>
#ifndef WIN32           //*NIX
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#include <signal.h>
#else                   //WIN32
#include <windows.h>
#include "resource.h"
#endif                  //ALL
#include <unistd.h>
#ifdef __DJGPP__
#define chroot(a) 0
#endif
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdlib.h>
#include <locale.h>
#include "path.h"
#include "partition.h"
#include "driver.h"
#include "arguments.h"

#include "crc8.h"
#include "compat.h"
#include "normal.h"
#include "log.h"
#include "message.h"
#include "buffer.h"
#ifdef __MINGW32__
#define mkdir(a, b) mkdir (a)
#endif

static Arguments arguments;
static Buffer buff[16];

#ifndef WIN32
#ifndef  __DJGPP__
#define FORKING
static int pipefd = -1;
#endif
#endif

#ifndef WIN32
static void deinit(void) {
    unsigned int b;
    for (b = 0; b < 16; b++) {
        Buffer *buffer = &buff[b];
        if (buffer->file != NULL) fclose(buffer->file);
        if (buffer->data != NULL) free(buffer->data);
    }
    for (b = 0; b < 256; b++) partition_set_path(b, NULL);
}

static const Driver *terminate_driver = NULL;
static void terminate(int x) {
    signal(x, SIG_DFL);
    if (terminate_driver != NULL) terminate_driver->shutdown();
    deinit();
#ifdef FORKING
    if (pipefd >= 0) close(pipefd);
#endif
    log_flush();
    raise(x);
}

static void setup_terminate(const Driver *driver) {
    terminate_driver = driver;
    signal(SIGTERM, terminate);
    signal(SIGINT, terminate);
}
#endif

#ifdef FORKING
static pid_t childpid = -1;

static void terminate2(int x) {
    signal(x, SIG_DFL);
    if (childpid > 0) kill(childpid, x);
    raise(x);
}
#endif

void seterror(Errorcode c, unsigned int i1) {
    const char *msg;
    int ret;

    switch (c) {
    case ER_OK: msg = "OK"; break;
    case ER_FILES_SCRATCHED: msg = "FILES SCRATCHED"; break;
    case ER_PARTITION_SELECTED: msg = "PARTITION SELECTED"; break;
    case ER_READ_ERROR: if (i1 == 0) msg = "READ ERROR"; else msg = "CRC ERROR"; i1 = 0; break;
    case ER_WRITE_PROTECT_ON: msg = "WRITE PROTECT ON"; break;
    case ER_ACCESS_DENIED: msg = "ACCESS DENIED"; break;
    case ER_UNKNOWN_COMMAND: msg = "UNKNOWN COMMAND"; break;
    case ER_SYNTAX_ERROR:
    case ER_INVALID_FILENAME:
    case ER_MISSING_FILENAME: msg = "SYNTAX ERROR"; break;
    case ER_PATH_NOT_FOUND: msg = "PATH NOT FOUND"; break;
    case ER_FRAME_ERROR: msg = "FRAME ERROR"; break;
    case ER_CRC_ERROR: msg = "CRC ERROR"; break;
    case ER_FILE_NOT_FOUND: msg = "FILE NOT FOUND"; break;
    case ER_FILE_EXISTS: msg = "FILE EXISTS"; break;
    case ER_FILE_TYPE_MISMATCH: msg = "FILE TYPE MISMATCH"; break;
    case ER_NO_CHANNEL: msg = "NO CHANNEL"; break;
    case ER_PARTITION_FULL: msg = "PARTITION FULL"; break;
    case ER_DOS_VERSION: msg = "IDESERV " VERSION2 " PCLINK"; break;
    case ER_SELECTED_PARTITION_ILLEGAL: msg = "SELECTED PARTITION ILLEGAL"; break;
    default: msg = "UNKNOWN ERROR"; break;
    }
    ret = sprintf((char *)buff[15].data, "%02d, %s,%03u,000,000,000", c, msg, i1);
    buff[15].size = ret >= 0 ? (unsigned int)ret : 0;
    buff[15].pointer = 0;
}

Errorcode errnotochannel15(int i) {
    switch (i) {
    case 0: seterror(ER_OK, 0); return ER_OK;
    case EISDIR:
    case ENOTDIR:
    case ENOENT: seterror(ER_FILE_NOT_FOUND, 0); return ER_FILE_TYPE_MISMATCH;
    case EEXIST: seterror(ER_FILE_EXISTS, 0); return ER_FILE_EXISTS;
    case EPERM:
    case EACCES: seterror(ER_ACCESS_DENIED, 0); return ER_ACCESS_DENIED;
    case ENOTEMPTY: seterror(ER_FILES_SCRATCHED, 0); return ER_FILES_SCRATCHED;
    case ENOSPC: seterror(ER_PARTITION_FULL, 0); return ER_PARTITION_FULL;
    case EROFS: seterror(ER_WRITE_PROTECT_ON, 0); return ER_WRITE_PROTECT_ON;
    case EIO: seterror(ER_READ_ERROR, 0); return ER_READ_ERROR;
    default: seterror(ER_UNKNOWN_ERROR, (unsigned int)i); return ER_OK;
    }
}

#ifdef FORKING
static int partition_select2(partition_t partition) {
    if (pipefd >= 0) {
        unsigned char msg[3];
        msg[0] = 2;
        msg[1] = (unsigned char)partition;
        msg[2] = 0;
        write(pipefd, msg, sizeof msg);
    }
    return partition_select(partition);
}

static void partition_set_path2(partition_t partition, const char *path) {
    if (pipefd >= 0) {
        unsigned char msg[3];
        size_t pathlen = strlen(path);
        if (pathlen < 256) {
            msg[0] = 1;
            msg[1] = (unsigned char)partition;
            msg[2] = (unsigned char)pathlen;
            write(pipefd, msg, sizeof msg);
            if (msg[2] != 0) write(pipefd, path, msg[2]);
        }
    }
    partition_set_path(partition, path);
}
#else
#define partition_set_path2 partition_set_path
#define partition_select2 partition_select
#endif

static void do_chdir(const Petscii *cmd) {
    Petscii name[17];
    const Petscii *outname;
    char outpath[1020];
    int found = 0;
    Errorcode status = ER_OK;
    partition_t outpart = 0;
    Directory_entry dirent;
    Directory *directory;

    if (!strchr((char *)cmd, ':')) {
        if (!strcmp((char *)cmd, "_")) cmd = (Petscii *)":_";
        else if (cmd[0] != '/') {
            status = ER_SYNTAX_ERROR; goto vege3;
        } else cmd++;
    }

    outname = resolv_path(cmd, outpath, &outpart, arguments.nameconversion);
    if (outname == NULL) goto vege2;

    convertfilename(outname, ',', name, NULL, NULL);

    if ((name[0] == '_' && name[1] == 0) || (name[0] == '.' && name[1] == '.' && name[2] == 0)) {
        char *o = strrchr(outpath, '/');
        if (o == NULL) o = outpath;
        *o = 0;
        found = 1;
        goto vege2;
    }
    if (name[0] == 0 || (name[0] == '.' && name[1] == 0)) {
        found = 1;
        goto vege2;
    }

    directory = directory_open(outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotochannel15(errno);
        goto vege3;
    }

    while (directory_read(directory, &dirent)) {
        if ((dirent.attrib & A_ANY) != A_DIR) continue;

        if (!matchname(dirent.name, name)) continue;

        strcpy(outpath, directory_path(directory));
        found = 1;
        break;
    }
    directory_close(directory);
vege2:
    if (found) {
        partition_set_path2(outpart, outpath);
    } else {
        status = ER_PATH_NOT_FOUND;
    }
vege3:
    if (arguments.verbose) log_printf("Command: Change directory \"%s\"", outpath[0] ? outpath : "/");
    seterror(status, 0);
}

static void do_mkdir(const Petscii *cmd) {
    char outpath[1020], lname[1000];
    int found = 0;
    Errorcode status = ER_OK;
    Petscii name[17];
    const Petscii *outname;
    Directory_entry dirent;
    Directory *directory;

    if (!strchr((char *)cmd, ':')) {
        status = ER_SYNTAX_ERROR; outpath[0] = 0; goto vege2;
    }

    outname = resolv_path(cmd, outpath, NULL, arguments.nameconversion);
    if (outname == NULL) {
        status = ER_PATH_NOT_FOUND; goto vege2;
    }

    convertfilename(outname, ',', name, NULL, NULL);

    if (!name[0]) {
        status = ER_MISSING_FILENAME;
        goto vege2;
    }

    directory = directory_open(outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotochannel15(errno);
        goto vege2;
    }

    while (directory_read(directory, &dirent))
    {
        if ((dirent.attrib & A_ANY) != A_DIR) continue;

        if (!matchname(dirent.name, name)) continue;

        strncpy(lname, directory_filename(directory), 999);

        found = 1;
        break;
    }
    directory_close(directory);

    if (!found) {
        mbstate_t ps;
        unsigned int f, f2;
        memset(&ps, 0, sizeof ps);
        for (f = f2 = 0; name[f] && f < 16; f++) {
            f2 += c64toascii(lname + f2, name[f], &ps);
        }
        lname[f2] = 0;
    }

    if (outpath[0]) strcat(outpath, "/");
    strcat(outpath, lname);

    if (found)
        status = ER_FILE_EXISTS;
    else {
        int f = 0;

        while (lname[f]) {
            if (lname[f] == '*' || lname[f] == '?' || lname[f] == ':' || lname[f] == '=') {
                status = ER_INVALID_FILENAME; goto vege2;
            }
            f++;
        }
        if (mkdir(outpath, 0777)) {
            status = errnotochannel15(errno);
        }
    }
vege2:
    if (arguments.verbose) log_printf("Command: Make directory \"%s\"", outpath[0] ? outpath : "/");
    seterror(status, 0);
}

static void do_scratch(const Petscii *cmd, int smode) {
    char outpath[1000];
    unsigned int found = 0;
    Errorcode status = ER_FILES_SCRATCHED;
    Petscii name[17], type[4] = {'*', 0};
    const Petscii *outname;
    Directory_entry dirent;
    Directory *directory;

    if (!strchr((char *)cmd, ':')) {
        status = ER_SYNTAX_ERROR; outpath[0] = 0; goto vege;
    }

    outname = resolv_path(cmd, outpath, NULL, arguments.nameconversion);
    if (outname == NULL) {
        status = ER_PATH_NOT_FOUND; goto vege;
    }

    convertfilename(outname, '=', name, type, NULL);

    if (!name[0]) {
        status = ER_MISSING_FILENAME;
        goto vege;
    }

    directory = directory_open(outpath, arguments.nameconversion, 0);
    if (directory == NULL) {
        status = errnotochannel15(errno);
        goto vege;
    }

    while (directory_read(directory, &dirent))
    {
        const char *path;
        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED)) != (A_DIR | A_CLOSED)) continue;
        } else {
            if ((dirent.attrib & A_ANY) == A_DIR) continue;
        }

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        if (!matchname(dirent.name, name)) continue;
        if (!smode) {
            if (!matchname(dirent.filetype, type)) continue;
        }

        path = directory_path(directory);
        found++;
        if (smode) {
            if (rmdir(path)) {
                status = errnotochannel15(errno);
            }
            if (arguments.verbose) log_printf("Command: Remove directory \"%s\"", path);
        } else {
            if (unlink(path)) {
                status = errnotochannel15(errno);
            }
            if (arguments.verbose) log_printf("Command: Remove \"%s\"", path);
        }
    }
    directory_close(directory);
vege:
    if (!found) {
        if (smode) {
            if (arguments.verbose) log_printf("Command: Remove directory \"%s\"", outpath);
        } else {
            if (arguments.verbose) log_printf("Command: Remove \"%s\"", outpath);
        }
    }
    seterror(status, (status == ER_FILES_SCRATCHED) ? found : 0);
}

void commandchannel(const Petscii *s) {
    if (s[0] == 'C' && s[1] == 'D') {
        do_chdir(s + 2);
    } else if (s[0] == 'M' && s[1] == 'D') {
        do_mkdir(s);
    } else if (s[0] == 'R' && s[1] == 'D') {
        do_scratch(s, 1);
    } else if (s[0] == 'S' && s[1] != '-') {
        do_scratch(s, 0);
    } else if (s[0] == 'C' && s[1] == 0xd0) {
        if (arguments.verbose) log_printf("Command: Change to partition %d", s[2]);
        seterror(partition_select2(s[2]) ? ER_SELECTED_PARTITION_ILLEGAL : ER_PARTITION_SELECTED, s[2]);
    } else if (s[0] == 'C' && s[1] == 'P') {
        partition_t part = 0;
        int i = 2;
        if (!s[2]) part = partition_get_current(); else {
            while ((s[i] ^ 0x30) < 10) part = part * 10 + (s[i++] ^ 0x30);
        }
        if (arguments.verbose) log_printf("Command: Change to partition %d", part);
        seterror(partition_select2(part) ? ER_SELECTED_PARTITION_ILLEGAL : ER_PARTITION_SELECTED, part);
    } else if (s[0] == 'U' && ((s[1] & 0x0f) == 9 || (s[1] & 0x0f) == 10)) {
        seterror(ER_DOS_VERSION, 0);
        if (arguments.verbose) log_print("Command: Identify device");
    } else if (s[0] == 'I') {
        seterror(ER_OK, 0);
        if (arguments.verbose) log_print("Command: Initialize");
    } else if (s[0] == 'V') {
        seterror(ER_OK, 0);
        if (arguments.verbose) log_print("Command: Validate");
    } else if (s[0] == 'M' && s[1] == '-' && s[2] == 'R') {
        int i, j, k;
        j = s[5]; if (j > 50) j = 50;
        k = s[3];
        for (i = 0; i < j; i++) buff[15].data[i] = (unsigned char)"IDE64 CARTRIDGE "[k++ & 15];
        buff[15].data[i] = 0;
        if (arguments.verbose) log_printf("Command: Memory read $%04x %02x", s[3] | (s[4] << 8), s[5]);
    } else if (s[0]) {
        seterror(ER_UNKNOWN_COMMAND, 0);
    }
}

typedef struct Environment {
    const char *root;
#ifndef WIN32
#ifndef __DJGPP__
    int setuid;
    uid_t uid;
    int setgid;
    gid_t gid;
#endif
#endif
} Environment;

static int prepare_environment(Environment *environment) {
    environment->root = arguments.root;
    if (arguments.root != NULL) {
        if (chdir(arguments.root) != 0) {
            message("Changing dir to \"%s\" failed: %s(%d)\n", arguments.root, strerror(errno), errno);
            return 1;
        }
    }
#ifndef WIN32
#ifndef __DJGPP__
    if (arguments.root != NULL) {
        time_t now = time(NULL);
        localtime(&now);
    }
    errno = 0;
    if (arguments.priority && nice(arguments.priority) == -1 && errno != 0) {
        message("Priority adjustment failed: %s(%d)\n", strerror(errno), errno);
        return 1;
    }
    if (arguments.user == NULL || arguments.group == NULL) {
        uid_t uid = getuid(), euid = geteuid();
        if (uid <= 0 || uid != euid) {
            if (arguments.user == NULL) arguments.user = "nobody";
            if (arguments.group == NULL) arguments.group = "nogroup";
        }
    }
    environment->setuid = arguments.user != NULL;
    if (environment->setuid) {
        struct passwd *user2;
        errno = 0;
        user2 = getpwnam(arguments.user);
        if (user2 == NULL) {
            if (errno == 0 || errno == ENOENT || errno == ESRCH || errno == EBADF || errno == EPERM) {
                message("No such user name: %s\n", arguments.user);
            } else {
                message("User name search failed: %s(%d)\n", strerror(errno), errno);
            }
            return 1;
        }
        environment->uid = user2->pw_uid;
    }
    environment->setgid = arguments.group != NULL;
    if (environment->setgid) {
        struct group *group2;
        errno = 0;
        group2 = getgrnam(arguments.group);
        if (group2 == NULL) {
            if (errno == 0 || errno == ENOENT || errno == ESRCH || errno == EBADF || errno == EPERM) {
                message("No such group name: %s\n", arguments.group);
            } else {
                message("Group name search failed: %s(%d)\n", strerror(errno), errno);
            }
            return 1;
        }
        environment->gid = group2->gr_gid;
    }
#endif
#endif
    return 0;
}

static int setup_environment(Environment *environment) {
    unsigned int b;
#ifdef WIN32
#ifdef __CYGWIN__
    if (environment->root == NULL) {
        if (chdir("/cygdrive") != 0) {
            log_printf("Changing dir to \"/cygdrive\" failed: %s(%d)", strerror(errno), errno);
        }
    }
#endif
#else
#ifndef __DJGPP__
    if (environment->root != NULL) {
        uid_t uid = getuid(), euid = geteuid();
        if (chroot(environment->root) != 0 || chdir("/") != 0) {
            log_printf("Chrooting failed: %s(%d)", strerror(errno), errno);
            if (uid <= 0 || uid != euid) {
                return 1;
            }
        }
    }
    if (environment->setgid) {
        gid_t gid;
        int groups = getgroups(0, NULL);
        if (groups != 0 && (groups != 1 || getgroups(1, &gid) == -1 || environment->gid != gid)) {
            if (setgroups(1, &environment->gid) != 0) {
                log_print("Could not drop all groups");
                return 1;
            }
        }
        if (setgid(environment->gid) != 0 || environment->gid != getgid() || environment->gid != getegid()) {
            log_printf("Could not change group: %s(%d)", strerror(errno), errno);
            return 1;
        }
        groups = getgroups(0, NULL);
        if (groups != 0 && (groups != 1 || getgroups(1, &gid) == -1 || environment->gid != gid)) {
            log_print("Could not drop all groups");
            return 1;
        }
    }
    if (environment->setuid) {
        if (setuid(environment->uid) != 0 || environment->uid != getuid() || environment->uid != geteuid()) {
            log_printf("Could not change user: %s(%d)", strerror(errno), errno);
            return 1;
        }
    }
#endif
#endif
    for (b = 0; b < 15; b++) buff[b].mode = CM_CLOSED;
    buff[15].mode = CM_ERR;
    if (buffer_reserve(&buff[15], 256)) {
        message("Out of memory");
        return 1;
    }
    seterror(ER_DOS_VERSION, 0);
    return 0;
}

static int rate_limit(struct timeval *start) {
    struct timeval end;
    if (gettimeofday(&end, NULL) == -1) return 1;
    if (end.tv_sec < start->tv_sec) return 1;
    if (start->tv_sec == end.tv_sec) {
        if (end.tv_usec < start->tv_usec) return 1;
        end.tv_usec -= start->tv_usec;
        if (end.tv_usec < 1000000) usleep(1000000 - end.tv_usec);
    } else if (start->tv_sec + 1 == end.tv_sec) {
        if (start->tv_usec > end.tv_usec) {
            end.tv_usec = start->tv_usec - end.tv_usec;
            if (end.tv_usec < 1000000) usleep(end.tv_usec);
        }
    }
    return 0;
}

static void driver_initialize(const Driver *driver) {
    int lastfail = 0;
    log_flush();
    for (;;) {
        struct timeval start;
        int ret = gettimeofday(&start, NULL);
        lastfail = driver->initialize(lastfail);
        if (lastfail == 0) {
            return;
        }
        log_flush();
        if (ret == -1 || rate_limit(&start)) {
            sleep(1);
        }
    }
}

static void serve(const Driver *driver) {
    int b;
    unsigned char ec = 0x5a;

    for (;;) {
        log_flush();
    start2:
        crc_clear(0);
        b = driver->wait(ec);
        if (b < 0) return;

        switch (b) {
        case 0x00: goto start2;
        case 'N': crc_clear(-1); /* fall through */
        case 0xCE: b = openfile(driver, buff, &arguments, b == 0xCE ? FLAG_USE_CRC : 0); ec = 0x5a; break;
        case 'G': crc_clear(-1); /* fall through */
        case 0xC7: b = readfile(driver, buff, &arguments, b == 0xC7 ? FLAG_USE_CRC : 0); ec = 0x5a; break;
        case 'P': crc_clear(-1); /* fall through */
        case 'S': crc_clear(-1); /* fall through */
        case 0xD3: b = writefile(driver, buff, &arguments, b == 0xD3 ? FLAG_USE_CRC : b == 'P' ? FLAG_USE_PADDING : 0); ec = 0x5a; break;
        case 'D': crc_clear(-1); /* fall through */
        case 0xC4: b = closefile(driver, buff, &arguments, b == 0xC4 ? FLAG_USE_CRC : 0); ec = 0x5a; break;
        case 'I': crc_clear(-1); /* fall through */
        case 0xC9: b = statuserror(driver, (char *)buff[15].data, &arguments, b == 0xC9 ? FLAG_USE_CRC : 0); ec = 0x5a; break;
        case 'O': crc_clear(-1); /* fall through */
        case 0xCF: b = openfile_compat(driver, buff, &arguments, b == 0xCF ? FLAG_USE_CRC : 0); ec = 0; break;
        case 'R': crc_clear(-1); /* fall through */
        case 0xD2: b = readfile_compat(driver, buff, &arguments, b == 0xD2 ? FLAG_USE_CRC : 0); ec = 0; break;
        case 'W': crc_clear(-1); /* fall through */
        case 0xD7: b = writefile_compat(driver, buff, &arguments, b == 0xD7 ? FLAG_USE_CRC : 0); ec = 0; break;
        case 'C': crc_clear(-1); /* fall through */
        case 0xC3: b = closefile_compat(driver, buff, &arguments, b == 0xC3 ? FLAG_USE_CRC : 0); ec = 0; break;
        default: log_printf("Unknown command %02X", b);
        }
        if (b != 0) {
            if (driver->clean()) log_print("Timeout");
        }
    }
}

#ifdef WIN32
static BOOL CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
    switch (message) {
    case WM_COMMAND:
        switch (wParam) {
        case IDOK:
        case IDCANCEL:
        case IDC_EXIT:
            DestroyWindow(hwnd);
            return TRUE;
        case IDC_HIDE:
            ShowWindow(hwnd, SW_HIDE);
            return TRUE;
        }
        break;
    case WM_USER + 1:
        {
            HWND logwindow_handle = GetDlgItem(hwnd, IDC_LOG);
            int len = GetWindowTextLength(logwindow_handle);
            static int max_size;
            if (max_size == 0) max_size = SendMessage(logwindow_handle, EM_GETLIMITTEXT, 0, 0);
            SendMessage(logwindow_handle, WM_SETREDRAW, FALSE, 0);
            while (len + 256 > max_size && max_size > 512) {
                SendMessage(logwindow_handle, EM_SETSEL, 0, SendMessage(logwindow_handle, EM_LINELENGTH, 0, 0) + 2);
                SendMessage(logwindow_handle, EM_REPLACESEL, 0, (LPARAM)"");
                len = GetWindowTextLength(logwindow_handle);
            }
            if (len != 0) {
                SendMessage(logwindow_handle, EM_SETSEL, len, len);
                SendMessage(logwindow_handle, EM_REPLACESEL, 0, (LPARAM)"\r\n");
                len += 2;
            }
            SendMessage(logwindow_handle, EM_REPLACESEL, 0, lParam);
            SendMessage(logwindow_handle, EM_SETSEL, len, len);
            SendMessage(logwindow_handle, WM_SETREDRAW, TRUE, 0);
            SendMessage(logwindow_handle, EM_SCROLLCARET, 0, 0);
            free((void *)lParam);
        }
        break;
    case WM_SIZE:
        {
            RECT r;
            int bottom_margin, top_margin;
            int width = LOWORD(lParam);
            int height = HIWORD(lParam);
            r.left = 0; r.top = 0; r.right = 170; r.bottom = 105;
            MapDialogRect(hwnd, &r);
            bottom_margin = r.bottom;
            top_margin = r.bottom;
            r.left = 0; r.top = 0; r.right = 170; r.bottom = 80;
            MapDialogRect(hwnd, &r);
            bottom_margin -= r.bottom;
            MoveWindow(GetDlgItem(hwnd, IDC_LOG), r.left, r.top, width, height - bottom_margin, TRUE);

            r.left = 5; r.top = 85; r.right = 5 + 95; r.bottom = 85 + 15;
            MapDialogRect(hwnd, &r);
            top_margin -= r.top;
            MoveWindow(GetDlgItem(hwnd, IDC_HIDE), r.left, height - top_margin, r.right - r.left, r.bottom - r.top, TRUE);
            r.left = 105; r.top = 85; r.right = 105 + 60; r.bottom = 85 + 15;
            MapDialogRect(hwnd, &r);
            MoveWindow(GetDlgItem(hwnd, IDC_EXIT), r.left, height - top_margin, r.right - r.left, r.bottom - r.top, TRUE);
        }
        break;
    case WM_USER:
        if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN) {
            ShowWindow(hwnd, SW_SHOW);
            return TRUE;
        }
        break;
    case WM_DESTROY:
        PostQuitMessage(0);
        return TRUE;
    }
    return FALSE;
}

static DWORD WINAPI PCLinkLoop(long lpParam) {
    const Driver *driver = (const Driver *)lpParam;
    for (;;) {
        driver_initialize(driver);
        serve(driver);
        driver->shutdown();
    }
    return 0;
}
#endif //WIN32

int main(int argc, char *argv[]) {
    const Driver *driver;
    Environment environment;
#ifdef WIN32
    HWND hwnd;
    NOTIFYICONDATA notifyIconData;
    BOOL icon;
    HANDLE hThread;
    DWORD iID;
#endif

    setlocale(LC_CTYPE, "");

#ifdef WIN32
    hwnd = CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    logwindow_set(hwnd);

    log_print("IDEDOS 0.9x PCLink ideservd " VERSION "\r\n"
    "Written by Kajtar Zsolt <soci@c64.rulez.org>\r\n"
    "Win32 port by J.Soucek <josef@ide64.org>\r\n"
    "http://idedos.ide64.org/\r\n"
    "This program is free software. See the GNU\r\n"
    "General Public License for more details.\r\n");
#endif

    testarg(&arguments, argc, argv);
    log_open(arguments.log);
    log_print("ideservd " VERSION);

    driver = get_driver(&arguments);
    if (driver == NULL) {
        message("No such driver: %s\n", arguments.mode_name);
        return EXIT_FAILURE;
    }

    if (prepare_environment(&environment)) return EXIT_FAILURE;

    partition_create(1, (Petscii *)"PARTITION 1");
    partition_select(1);

#ifdef WIN32
    if (setup_environment(&environment)) return EXIT_FAILURE;
    if (arguments.background != 1) ShowWindow(hwnd, SW_SHOW);
    memset(&notifyIconData, 0, sizeof notifyIconData);
    notifyIconData.hIcon = (HICON)LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON));
    if (notifyIconData.hIcon == NULL) {
        icon = FALSE;
    } else {
        SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)notifyIconData.hIcon);
        notifyIconData.cbSize = sizeof notifyIconData;
        notifyIconData.hWnd = hwnd;
        notifyIconData.uID = 0;
        notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
        notifyIconData.uCallbackMessage = WM_USER;
        strcpy(notifyIconData.szTip, "IDE64 PCLink");
        icon = Shell_NotifyIcon(NIM_ADD, &notifyIconData);
    }

    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)PCLinkLoop, (LPVOID)driver, 0, &iID);
    if (hThread != NULL) {
        MSG msg;
        while (GetMessage(&msg, NULL, 0, 0) != 0) {
            if (!IsDialogMessage(hwnd, &msg)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
        TerminateThread(hThread, 0);
        CloseHandle(hThread);
    }
    if (icon) {
        Shell_NotifyIcon(NIM_DELETE, &notifyIconData);
    }
    logwindow_set(NULL);
#elif defined FORKING
    if (arguments.background == 1) {
        pid_t pid = fork();
        if (pid == -1) {
            message("Could not fork to background: %s(%d)\n", strerror(errno), errno);
            return EXIT_FAILURE;
        }
        if (pid != 0) {
            message("Forking into background\n");
            return EXIT_SUCCESS;
        }
        if (setsid() == -1) {
            message("Setsid failed: %s(%d)\n", strerror(errno), errno);
            return EXIT_FAILURE;
        }
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
    }
    if (arguments.background >= 0) {
        signal(SIGTERM, terminate2);
        signal(SIGINT, terminate2);
        for (;;) {
            struct timeval start;
            int ret;
            pid_t pid;
            int status;
            int pipefds[2];
            if (pipe(pipefds) == -1) {
                pipefds[0] = -1;
                pipefds[1] = -1;
            }
            pid = fork();
            if (pid == -1) {
                log_print("Fork failed");
                return EXIT_FAILURE;
            }
            if (pid == 0) {
                if (pipefds[0] >= 0) close(pipefds[0]);
                pipefd = pipefds[1];
                setup_terminate(driver);
                driver_initialize(driver);
                if (!setup_environment(&environment)) serve(driver);
                driver->shutdown();
                deinit();
                if (pipefds[1] >= 0) close(pipefds[1]);
                return EXIT_SUCCESS;
            }
            childpid = pid;
            ret = gettimeofday(&start, NULL);
            if (pipefds[1] >= 0) close(pipefds[1]);
            if (pipefds[0] >= 0) {
                unsigned char msg[3];
                while (read(pipefds[0], &msg, sizeof msg) == sizeof msg) {
                    if (msg[0] == 1) {
                        char path[256];
                        ssize_t c = (msg[2] != 0) ? read(pipefds[0], path, msg[2]) : 0;
                        if (c != msg[2]) break;
                        path[c] = 0;
                        partition_set_path(msg[1], path);
                    } else if (msg[0] == 2) {
                        partition_select(msg[1]);
                    } else {
                        break;
                    }
                }
                close(pipefds[0]);
            }
            if (waitpid(pid, &status, 0) == -1) {
                log_print("Waitpid failed");
                return EXIT_FAILURE;
            }
            childpid = -1;
            if (WIFSIGNALED(status)) {
                int sig = WTERMSIG(status);
                log_printf("Child caught signal %d", sig);
                raise(sig);
            }
            if (ret == -1 || rate_limit(&start)) {
                usleep(1);
            }
        }
    } else {
        setup_terminate(driver);
        driver_initialize(driver);
        if (setup_environment(&environment)) return EXIT_FAILURE;
        serve(driver);
        driver->shutdown();
        deinit();
    }
#else
#ifndef __DJGPP__
    if (arguments.background == 1) {
        message("Not possible to run in background\n");
        return EXIT_FAILURE;
    }
#endif
    setup_terminate(driver);
    if (setup_environment(&environment)) return EXIT_FAILURE;
    for (;;) {
        driver_initialize(driver);
        serve(driver);
        driver->shutdown();
    }
#endif 
    return EXIT_SUCCESS;
}
