/*

 rs232.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "rs232.h"
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
static HANDLE commhandle = INVALID_HANDLE_VALUE;
static COMMTIMEOUTS commtimeouts;
#else
#include <sys/ioctl.h>
#include <termios.h>
static int fd = -1;
static struct termios options;
#ifndef FNDELAY
#define FNDELAY 0
#endif
#endif

#include "crc8.h"
#include "log.h"
#include "driver.h"

#if defined WIN32 || defined __DJGPP__
#define DEFAULT_COMPORT "COM1"
#else
#define DEFAULT_COMPORT "/dev/ttyS0"
#endif

static const char *i_dev;
static int i_baudlow;
static int inited;
static int driver_errno = -EIO;
static int timeout_expired;

#ifdef WIN32
static const char *errormsg(int err) {
    static char msg[100];
    DWORD len = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, (DWORD)err, MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT), (LPSTR)&msg, sizeof msg, NULL);
    if (len > 0 && msg[len - 1] == '\n') len--;
    if (len > 0 && msg[len - 1] == '\r') len--;
    if (len > 0 && msg[len - 1] == '.') len--;
    msg[len] = 0;
    return msg;
}
#endif

static int initialize(int lastfail) {
#ifdef WIN32
    DCB dcb;

    driver_errno = -EIO;
    inited = 0;
// Open the serial port
    if (commhandle == INVALID_HANDLE_VALUE) {
        commhandle = CreateFile(i_dev, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
    }
    if (commhandle == INVALID_HANDLE_VALUE) {
        if (lastfail != -1) {
            DWORD error = GetLastError();
            log_printf("Unable to open serial port: %s(%lu)", errormsg(error), error);
        }
        return -1;
    }
// Initialize the DCBlength member
    dcb.DCBlength = sizeof dcb;
// Get the default port setting information
    if (!GetCommState(commhandle, &dcb)) {
        if (lastfail != -2) {
            DWORD error = GetLastError();
            log_printf("Unable to get configuration of serial port: %s(%lu)", errormsg(error), error);
        }
    }
//Change the DCB structure settings
    dcb.BaudRate = i_baudlow ? 38400 : 115200;// Current baud
    dcb.fBinary = TRUE;               	// Binary mode, (windows supports only binary)
    dcb.fParity = TRUE;               	// Enable parity checking 
    dcb.fOutxCtsFlow = TRUE;         	// CTS output flow control 
    dcb.fOutxDsrFlow = FALSE;         	// No DSR output flow control 
    dcb.fDtrControl = DTR_CONTROL_ENABLE; 	// Enable DTR signal while device open
    dcb.fDsrSensitivity = FALSE;      	// DSR sensitivity 
    dcb.fTXContinueOnXoff = TRUE;     	// XOFF continues Tx 
    dcb.fOutX = FALSE;                	// No XON/XOFF out flow control 
    dcb.fInX = FALSE;                 	// No XON/XOFF in flow control 
    dcb.fErrorChar = FALSE;           	// Disable error replacement 
    dcb.fNull = FALSE;                	// Disable null stripping 
    dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;// RTS flow control 
    dcb.fAbortOnError = FALSE;        	// Do not abort reads/writes on error
    dcb.ByteSize = 8;                 	// Number of bits/byte, 4-8 
    dcb.Parity = NOPARITY;            	// 0-4=no,odd,even,mark,space 
    dcb.StopBits = ONESTOPBIT;        	// 0,1,2 = 1, 1.5, 2 
// Configure the port according to the specifications of the DCB structure
    if (!SetCommState(commhandle, &dcb)) {
        if (lastfail != -3) {
            DWORD error = GetLastError();
            log_printf("Unable to configure serial port: %s(%lu)", errormsg(error), error);
        }
        return -3;
    }
// Retrieve the timeout parameters for all read and write operations on the port
    if (!GetCommTimeouts(commhandle, &commtimeouts)) {
        if (lastfail != -4) {
            DWORD error = GetLastError();
            log_printf("Unable to get the timeout parameters: %s(%lu)", errormsg(error), error);
        }
        return -4;
    }

// Change the COMMTIMEOUTS structure settings
    commtimeouts.ReadIntervalTimeout = MAXDWORD;
    commtimeouts.ReadTotalTimeoutMultiplier = 0;
    commtimeouts.ReadTotalTimeoutConstant = 1000;       // MAXDWORD, 0, after x miliseconds read returns even if no bytes received
    commtimeouts.WriteTotalTimeoutMultiplier = 0;
    commtimeouts.WriteTotalTimeoutConstant = 1000;

// Set the timeout parameters for all read and write operations on the port
    if (!SetCommTimeouts(commhandle, &commtimeouts)) {
        if (lastfail != -5) {
            DWORD error = GetLastError();
            log_printf("Unable to set the timeout parameters: %s(%lu)", errormsg(error), error);
        }
        return -5;
    }
#else
    int flags;

    driver_errno = -EIO;
    inited = 0;
    /// Changes by Silver Dream !
    if (fd == -1) {
        fd = open(i_dev, O_RDWR | O_NOCTTY | O_NONBLOCK);
    }
    if (fd == -1) {
        if (lastfail != -1) {
            log_printf("Cannot open %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -1;
    }
#ifdef TIOCEXCL
    if (ioctl(fd, TIOCEXCL) == -1) {
        if (lastfail != -2) {
            log_printf("Error setting TIOCEXCL on %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -2;
    }
#endif
    if (fcntl(fd, F_SETFL, 0) == -1) {
        if (lastfail != -3) {
            log_printf("Error clearing O_NONBLOCK on %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -3;
    }

    if (tcgetattr(fd, &options) == -1) {
        if (lastfail != -4) {
            log_printf("Cannot read params %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -4;
    }
    // End of changes by Silver Dream !

    /* baud rate */
#ifdef B115200
    if (!i_baudlow) {
        cfsetispeed(&options, B115200);
        cfsetospeed(&options, B115200);
    } else
#endif
    {
        cfsetispeed(&options, B38400);
        cfsetospeed(&options, B38400);
    }

    /* input options */

    options.c_iflag = 0;

    /* cout options */

    options.c_oflag = 0;

    /* control options */
    /* 8N1 */

    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~CSTOPB;
    options.c_cflag &= ~CSIZE;
    options.c_cflag |= CS8;

    /* CTS/RTS flow control */

//    if (flowctrl)
//    {
//	printf("Enabling RTS/CTS flow control\n");
#ifdef CRTSCTS
    options.c_cflag |= CRTSCTS;
#endif
//    } else options.c_cflag &= ~CRTSCTS;

    /* line options */
    /* raw input */

//  options.c_lflag |= IEXTEN; /* leave enabled to get flow control */
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ISIG);

    /* control character */

//  options.c_cc[VMIN] = 1; /* minimum number of character to read */
    options.c_cc[VMIN] = 0; /* minimum number of character to read */
    options.c_cc[VTIME] = 10; /* time to wait for data */

    if (tcsetattr(fd, TCSANOW, &options) == -1) {
        if (lastfail != -5) {
            log_printf("Cannot set params on %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -5;
    }

    /* turn turn O_NONBLOCK again for sure */
    /* O_NONBLOCK = O_NDELAY -> DOESNT CARE WHAT STATE DCD */


    flags = fcntl(fd, F_GETFL, 0);
    if (flags == -1) {
        if (lastfail != -6) {
            log_printf("Error executing fcntl on %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -6;
    }
    if (fcntl(fd, F_SETFL, flags & ~(O_NONBLOCK | FNDELAY)) == -1) {
        if (lastfail != -7) {
            log_printf("Error executing fcntl on %s: %s(%d)", i_dev, strerror(errno), errno);
        }
        return -7;
    }

//fcntl(fd, F_SETFL, (flags | O_NONBLOCK) & ~FNDELAY);
/* set the read function to return immediately */

//  fcntl(fd, F_SETFL, FNDELAY);
#endif
    inited = 1;
    return 0;
}

static int getb(int use_timeout) {
    (void)use_timeout;
    unsigned char data;
#ifndef WIN32
    int err;
#else
    BOOL ok;
    DWORD bytesread;
#endif

    if (driver_errno != 0) return EOF;
#ifndef WIN32
    err = read(fd, &data, 1);
    if (err == 1) {
        crc_add_byte(data);
        return (unsigned char)data;
    }
#else
    ok = ReadFile(commhandle, &data, 1, &bytesread, NULL);
    if (ok) {
        if (bytesread == 1) {
            crc_add_byte(data);
            return (unsigned char)data;
        }
        timeout_expired = 1;
    }
#endif
    driver_errno = -EIO;
    return EOF;
}

static void sendb(unsigned char a) {
#ifndef WIN32
    int err;
#else
    DWORD byteswritten;
    BOOL ok;
#endif
    if (driver_errno != 0) return;
    crc_add_byte(a);
#ifndef WIN32
    do {
        err = write(fd, &a, 1);
        if (err < 0) {
            driver_errno = -EIO;
            return;
        }
    } while (err != 1);
#else
    ok = WriteFile(commhandle, &a, 1, &byteswritten, NULL);
    if (!ok || byteswritten != 1) {
        driver_errno = -EIO;
        return;
    }
#endif
}

static void getbytes(unsigned char data[], unsigned int bytes) {
    if (driver_errno != 0) return;
    do {
#ifndef WIN32
        unsigned int l;
        int err = read(fd, data, bytes);
        if (err <= 0) {
            driver_errno = -EIO;
            return;
        }
        l = (unsigned int)err;
        crc_add_block(data, l);
        bytes -= l; data += l;
#else
        DWORD bytesread;
        BOOL ok = ReadFile(commhandle, data, bytes, &bytesread, NULL);
        if (!ok) {
            driver_errno = -EIO;
            return;
        }
        if (bytesread == 0) {
            timeout_expired = 1;
            driver_errno = -EIO;
            return;
        }
        crc_add_block(data, bytesread);
        bytes -= bytesread; data += bytesread;
#endif
    } while (bytes > 0);
}

static void sendbytes(const unsigned char data[], unsigned int bytes) {
    if (driver_errno != 0) return;
    crc_add_block(data, bytes);
    do {
#ifndef WIN32
        unsigned int l;
        int err = write(fd, data, bytes);
        if (err < 0) {
            driver_errno = -EIO;
            return;
        }
        l = (unsigned int)err;
        bytes -= l; data += l;
#else
        DWORD byteswritten;
        BOOL ok = WriteFile(commhandle, data, bytes, &byteswritten, NULL);
        if (!ok || !byteswritten) {
            driver_errno = -EIO;
            return;
        }
        bytes -= byteswritten; data += byteswritten;
#endif
    } while (bytes > 0);
}

static void eshutdown(void) {
#ifndef WIN32
    if (fd != -1) {
        tcflush(fd, TCIOFLUSH);
        close(fd);
        fd = -1;
    }
#else
    if (commhandle != INVALID_HANDLE_VALUE) {
        CancelIo(commhandle);
        CloseHandle(commhandle);
        commhandle = INVALID_HANDLE_VALUE;
    }
#endif
    inited = 0;
    return;
}

static int flush(void) {
    int i;
    if (driver_errno == 0) {
#ifndef WIN32
        i = tcdrain(fd);
#else
        i = !FlushFileBuffers(commhandle);
#endif
        if (i != 0) return -EIO;
    }
    return driver_errno;
}

static int done(void) {
    return driver_errno;
}

static void turn(void) {
}

static int clean(void) {
    if (driver_errno == 0) {
#ifndef WIN32
        tcflush(fd, TCIOFLUSH);
#else
        CancelIo(commhandle);
#endif
    }
    return timeout_expired;
}

static int waitb(unsigned char ec) {
    int b;
    (void)ec;
    if (!inited) return -ENODEV;
#ifdef WIN32
    commtimeouts.ReadIntervalTimeout = 0;
    if (!SetCommTimeouts(commhandle, &commtimeouts)) {
        driver_errno = -ENODEV;
    } else {
        driver_errno = 0;
        b = getb(1);
        commtimeouts.ReadIntervalTimeout = MAXDWORD;
        if (!SetCommTimeouts(commhandle, &commtimeouts)) {
            driver_errno = -ENODEV;
        } else {
            timeout_expired = 0;
            if (b != EOF) return b;
        }
    }
#else
    options.c_cc[VMIN] = 1;
    if (tcsetattr(fd, TCSANOW, &options) == -1) {
        driver_errno = -ENODEV;
    } else {
        driver_errno = 0;
        b = getb(1);
        options.c_cc[VMIN] = 0;
        if (tcsetattr(fd, TCSANOW, &options) == -1) {
            driver_errno = -ENODEV;
        } else {
            timeout_expired = 0;
            if (b != EOF) return b;
        }
    }
#endif
    return driver_errno;
}

static Driver driver = {
    .name         = "RS232",
    .initialize   = initialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .done         = done,
    .turn         = turn,
    .wait         = waitb,
    .clean        = clean,
};

static const Driver *rs232_driver_generic(const char *dev, int baudlow) {
    i_dev = dev != NULL ? dev : DEFAULT_COMPORT;
    i_baudlow = baudlow;
    log_printf("Using %s driver on device %s", driver.name, i_dev);
    return &driver;
}

const Driver *rs232_driver(const char *dev) {
    driver.name = "RS232";
    return rs232_driver_generic(dev, 0);
}

const Driver *rs232s_driver(const char *dev) {
    driver.name = "RS232S";
    return rs232_driver_generic(dev, 1);
}
